//
//  test_functional.h
//  skiplist
//
//  Created by Paul Ross on 13/07/2016.
//  Copyright (c) 2016 Paul Ross. All rights reserved.
//

#ifndef __skiplist__test_functional__
#define __skiplist__test_functional__

int test_functional_all();

#endif /* defined(__skiplist__test_functional__) */
