//
//  test_print.h
//  skiplist
//
//  Created by Paul Ross on 13/07/2016.
//  Copyright (c) 2016 Paul Ross. All rights reserved.
//

#ifndef __skiplist__test_print__
#define __skiplist__test_print__

extern const int FUNCTION_WIDTH;
int print_result(const char *fn_name, int result);

#endif /* defined(__skiplist__test_print__) */
