//
//  test_rolling_median.h
//  skiplist
//
//  Created by Paul Ross on 13/07/2016.
//  Copyright (c) 2016 Paul Ross. All rights reserved.
//

#ifndef __skiplist__test_rolling_median__
#define __skiplist__test_rolling_median__

int test_rolling_median_all();

#endif /* defined(__skiplist__test_rolling_median__) */
